﻿// NativeCpuSkinning.cpp
#include "pch.h"
#include <malloc.h>
#include "SkinnedModel.h"
#include "Data/Matrix.h"
#include <vector>
#include <collection.h>

using namespace DirectX;
using namespace Platform;
using namespace Platform::Collections;
using namespace NativeHelper;

NativeHelper::SkinnedModel::SkinnedModel(void)
{
	_skinVertices = nullptr;

	_skinVerticesPositions = nullptr;
	_skinVerticesNormals = nullptr;
}

NativeHelper::SkinnedModel::~SkinnedModel(void)
{
	delete _skinVertices;

	_mm_free(_skinVerticesPositions);
	_mm_free(_skinVerticesNormals);
}

void NativeHelper::SkinnedModel::SetSkinVertices(const Array<SkinVertex>^ skinVertices)
{
	_verticesLength = skinVertices->Length;
	_skinVertices = new SkinVertex[_verticesLength];
	for(int i=0;i<_verticesLength;i++)
	{
		_skinVertices[i] = skinVertices[i];
	}

	_skinVerticesPositions = (XMVECTOR*)_mm_malloc(_verticesLength*sizeof(XMVECTOR), 16);
	_skinVerticesNormals = (XMVECTOR*)_mm_malloc(_verticesLength*sizeof(XMVECTOR), 16);
	for(int i=0;i<_verticesLength;i++)
	{		
		_skinVerticesPositions[i] = XMLoadFloat3((XMFLOAT3*)&skinVertices[i].Position);
		_skinVerticesNormals[i] = XMLoadFloat3((XMFLOAT3*)&skinVertices[i].Normal);
	}

	return;
}

void NativeHelper::SkinnedModel::Skin(const Array<Matrix3x4Data>^ bones, Platform::WriteOnlyArray<VertexPositionNormalTextureData>^ vertices)
{	
	// copy data locally
	int bonesLength = bones->Length;
	Matrix3x4Data* locbones = bones->Data;
	
	VertexPositionNormalTextureData* vout = vertices->Data;

	// skin all of the vertices
	int icount = _verticesLength;
	for (int i = 0; i < icount; i++)
	{	
		int b0 = _skinVertices[i].BlendIndices.X;
		int b1 = _skinVertices[i].BlendIndices.Y;
		int b2 = _skinVertices[i].BlendIndices.Z;
		int b3 = _skinVertices[i].BlendIndices.W;

		Matrix3x4Data* m1 = &locbones[b0];
		Matrix3x4Data* m2 = &locbones[b1];
		Matrix3x4Data* m3 = &locbones[b2];
		Matrix3x4Data* m4 = &locbones[b3];

		float w1 = _skinVertices[i].BlendWeights.X;
		float w2 = _skinVertices[i].BlendWeights.Y;
		float w3 = _skinVertices[i].BlendWeights.Z;
		float w4 = _skinVertices[i].BlendWeights.W;
		
		Matrix3x4Data skinnedTransformSum;
		skinnedTransformSum.M11 = (m1->M11 * w1) + (m2->M11 * w2) + (m3->M11 * w3) + (m4->M11 * w4);
		skinnedTransformSum.M12 = (m1->M12 * w1) + (m2->M12 * w2) + (m3->M12 * w3) + (m4->M12 * w4);
		skinnedTransformSum.M13 = (m1->M13 * w1) + (m2->M13 * w2) + (m3->M13 * w3) + (m4->M13 * w4);
		skinnedTransformSum.M21 = (m1->M21 * w1) + (m2->M21 * w2) + (m3->M21 * w3) + (m4->M21 * w4);
		skinnedTransformSum.M22 = (m1->M22 * w1) + (m2->M22 * w2) + (m3->M22 * w3) + (m4->M22 * w4);
		skinnedTransformSum.M23 = (m1->M23 * w1) + (m2->M23 * w2) + (m3->M23 * w3) + (m4->M23 * w4);
		skinnedTransformSum.M31 = (m1->M31 * w1) + (m2->M31 * w2) + (m3->M31 * w3) + (m4->M31 * w4);
		skinnedTransformSum.M32 = (m1->M32 * w1) + (m2->M32 * w2) + (m3->M32 * w3) + (m4->M32 * w4);
		skinnedTransformSum.M33 = (m1->M33 * w1) + (m2->M33 * w2) + (m3->M33 * w3) + (m4->M33 * w4);
		skinnedTransformSum.M41 = (m1->M41 * w1) + (m2->M41 * w2) + (m3->M41 * w3) + (m4->M41 * w4);
		skinnedTransformSum.M42 = (m1->M42 * w1) + (m2->M42 * w2) + (m3->M42 * w3) + (m4->M42 * w4);
		skinnedTransformSum.M43 = (m1->M43 * w1) + (m2->M43 * w2) + (m3->M43 * w3) + (m4->M43 * w4);
		
		// Support the 4 Bone Influences - Position then Normal
		Vector3Data position = _skinVertices[i].Position;
		vout[i].Position.X = position.X * skinnedTransformSum.M11 + position.Y * skinnedTransformSum.M21 + position.Z * skinnedTransformSum.M31 + skinnedTransformSum.M41;
		vout[i].Position.Y = position.X * skinnedTransformSum.M12 + position.Y * skinnedTransformSum.M22 + position.Z * skinnedTransformSum.M32 + skinnedTransformSum.M42;
        vout[i].Position.Z = position.X * skinnedTransformSum.M13 + position.Y * skinnedTransformSum.M23 + position.Z * skinnedTransformSum.M33 + skinnedTransformSum.M43;
		Vector3Data normal = _skinVertices[i].Normal;
		vout[i].Normal.X = normal.X * skinnedTransformSum.M11 + normal.Y * skinnedTransformSum.M21 + normal.Z * skinnedTransformSum.M31;
		vout[i].Normal.Y = normal.X * skinnedTransformSum.M12 + normal.Y * skinnedTransformSum.M22 + normal.Z * skinnedTransformSum.M32;
		vout[i].Normal.Z = normal.X * skinnedTransformSum.M13 + normal.Y * skinnedTransformSum.M23 + normal.Z * skinnedTransformSum.M33;

		vout[i].TextureCoordinate = _skinVertices[i].TextureCoordinate;
	}
	
	return;
}

void NativeHelper::SkinnedModel::SkinParral(const Array<Matrix3x4Data>^ bones, Platform::WriteOnlyArray<VertexPositionNormalTextureData>^ vertices)
{	
	// copy data locally
	int bonesLength = bones->Length;
	Matrix3x4Data* locbones = bones->Data;
	
	VertexPositionNormalTextureData* vout = vertices->Data;

	#pragma loop(hint_parallel(0))
	#pragma loop(ivdep)
	for (int i = 0, imax = _verticesLength; i < imax; i++)
	{
		int b0 = _skinVertices[i].BlendIndices.X;
		int b1 = _skinVertices[i].BlendIndices.Y;
		int b2 = _skinVertices[i].BlendIndices.Z;
		int b3 = _skinVertices[i].BlendIndices.W;

		Matrix3x4Data* m1 = &locbones[b0];
		Matrix3x4Data* m2 = &locbones[b1];
		Matrix3x4Data* m3 = &locbones[b2];
		Matrix3x4Data* m4 = &locbones[b3];
						
		float w1 = _skinVertices[i].BlendWeights.X;
		float w2 = _skinVertices[i].BlendWeights.Y;
		float w3 = _skinVertices[i].BlendWeights.Z;
		float w4 = _skinVertices[i].BlendWeights.W;

		Matrix3x4Data skinnedTransformSum;
		skinnedTransformSum.M11 = (m1->M11 * w1) + (m2->M11 * w2) + (m3->M11 * w3) + (m4->M11 * w4);
		skinnedTransformSum.M12 = (m1->M12 * w1) + (m2->M12 * w2) + (m3->M12 * w3) + (m4->M12 * w4);
		skinnedTransformSum.M13 = (m1->M13 * w1) + (m2->M13 * w2) + (m3->M13 * w3) + (m4->M13 * w4);
		skinnedTransformSum.M21 = (m1->M21 * w1) + (m2->M21 * w2) + (m3->M21 * w3) + (m4->M21 * w4);
		skinnedTransformSum.M22 = (m1->M22 * w1) + (m2->M22 * w2) + (m3->M22 * w3) + (m4->M22 * w4);
		skinnedTransformSum.M23 = (m1->M23 * w1) + (m2->M23 * w2) + (m3->M23 * w3) + (m4->M23 * w4);
		skinnedTransformSum.M31 = (m1->M31 * w1) + (m2->M31 * w2) + (m3->M31 * w3) + (m4->M31 * w4);
		skinnedTransformSum.M32 = (m1->M32 * w1) + (m2->M32 * w2) + (m3->M32 * w3) + (m4->M32 * w4);
		skinnedTransformSum.M33 = (m1->M33 * w1) + (m2->M33 * w2) + (m3->M33 * w3) + (m4->M33 * w4);
		skinnedTransformSum.M41 = (m1->M41 * w1) + (m2->M41 * w2) + (m3->M41 * w3) + (m4->M41 * w4);
		skinnedTransformSum.M42 = (m1->M42 * w1) + (m2->M42 * w2) + (m3->M42 * w3) + (m4->M42 * w4);
		skinnedTransformSum.M43 = (m1->M43 * w1) + (m2->M43 * w2) + (m3->M43 * w3) + (m4->M43 * w4);
		
		// Support the 4 Bone Influences - Position then Normal
		Vector3Data position = _skinVertices[i].Position;
		vout[i].Position.X = position.X * skinnedTransformSum.M11 + position.Y * skinnedTransformSum.M21 + position.Z * skinnedTransformSum.M31 + skinnedTransformSum.M41;
		vout[i].Position.Y = position.X * skinnedTransformSum.M12 + position.Y * skinnedTransformSum.M22 + position.Z * skinnedTransformSum.M32 + skinnedTransformSum.M42;
		vout[i].Position.Z = position.X * skinnedTransformSum.M13 + position.Y * skinnedTransformSum.M23 + position.Z * skinnedTransformSum.M33 + skinnedTransformSum.M43;
		Vector3Data normal = _skinVertices[i].Normal;
		vout[i].Normal.X = normal.X * skinnedTransformSum.M11 + normal.Y * skinnedTransformSum.M21 + normal.Z * skinnedTransformSum.M31;
		vout[i].Normal.Y = normal.X * skinnedTransformSum.M12 + normal.Y * skinnedTransformSum.M22 + normal.Z * skinnedTransformSum.M32;
		vout[i].Normal.Z = normal.X * skinnedTransformSum.M13 + normal.Y * skinnedTransformSum.M23 + normal.Z * skinnedTransformSum.M33;
		
		vout[i].TextureCoordinate = _skinVertices[i].TextureCoordinate;
	}
	
	return;
}
